// Title:	BiDiPro DLL module
// Desc:	SPI flash
// Ver:		0.1.3
// Found:	2008-09-26
// Last:	2011-06-25
// CPU:		386+
// Model:	Tiny
// Compil:	Borland C++ 3.1 (DOS)
// Author:	Oleg O. Chukaev aka 0leg


/****
25F BiDiPro module
Copyright (c) 2008-2011, 㪠   <oleg.chukaev@mail.ru>

蠥 ୮ ࠭  ᯮ짮    室
, ⠪   筮 ଥ,  ﬨ  ,  ᮡ ᫥
᫮:

1.  ୮ ࠭ 室   ⠢ 㪠
    㢥  ᪮ ࠢ,  ᯨ᮪ ᫮  ᫥騩
   ⪠  ࠭⨩.
2.  ୮ ࠭ 筮   ࠭ 㪠
    ଠ  ᪮ ࠢ,  ᯨ᮪ ᫮  ᫥騩
   ⪠  ࠭⨩  㬥樨 /  㣨 ਠ, ⠢塞
    ࠭.

      /  
"  "  -  ,   
, ,    ,  
      .    ,
    ,      ,
         ,   
/   ,    ,  
,   , ,   
,      
(,     ,  , 
,    -    , 
      ),   
         .
****/


#include <stdlib.h>
#include <limits.h>
#include "../dlldata.h"
#include "../errors.h"
#include "../bidi.h"


#ifdef LINUXXX
 extern DLLDATA *dlp;
#else
# define dlp ((DLLDATA *) NULL)		/* ⥫    */
#endif

#define bdport dlp->bdprt		/* ⥫  hardware ࠩ */

typedef struct CHDEFS {
	dword size;	// 𬪮
} CHDEFS;

CHDEFS chd[]={
//	size
	{0x000001L},		//1
	{0x008000L},		//256
	{0x010000L},		//512
	{0x020000L},		//010
	{0x040000L},		//020
	{0x080000L},		//040
	{0x100000L},		//080
	{0x200000L},		//016
	{0x400000L},		//032
};

// ७   CFG:
static struct CFGKEY cfgkeys[] = {
	{'D', "id_alg", 0, ULONG_MAX, 20},		//Read ID mode
	{'D', "wrsr_alg", 0, ULONG_MAX, 20},		//WRSR alg.
	{'D', "er_alg", 0, ULONG_MAX, 20},		//Erase mode
	{'D', "rd_alg", 0, ULONG_MAX, 20},		//Read mode
	{'D', "wr_alg", 1, ULONG_MAX, 20},		//Write mode (0..4)
	{'H', "flags", 0, ULONG_MAX, 0xffffL},		//Flags
	{'D', "wpage", 0, ULONG_MAX, 65536L},		//Page size
	{'D', "wtime", 20, ULONG_MAX, 100000L},		//Write time (us)
	{'D', "etime", 100000L, ULONG_MAX, 40000000L},	//Erase time (us)
	{0, {0}, 0, 0},					//End of table
};

#define cfg_id_alg	cfgkeys[0].data
#define cfg_wrsr_alg	cfgkeys[1].data
#define cfg_er_alg	cfgkeys[2].data
#define cfg_rd_alg	cfgkeys[3].data
#define cfg_wr_alg	cfgkeys[4].data
#define cfg_flags	cfgkeys[5].data
#define cfg_wpage	cfgkeys[6].data
#define cfg_wtime	cfgkeys[7].data
#define cfg_etime	cfgkeys[8].data

/* Output, 1B */
#define PIN_CE		0x01
#define PIN_WP		0x04
#define PIN_SI		0x08
#define PIN_SCK		0x10
#define PIN_HOLD	0x20

/* Output, 1C */
#define PIN_SST45_WP	0x01

/* Input, 1A */
#define PIN_SO		0x01

#define SPI_READ	0x03
#define SPI_SERASE	0x20
#define SPI_BERASE	0x52
#define SPI_CERASE	0x60
#define SPI_CERASE1	0x62
#define SPI_CERASE2	0xc7
#define SPI_PRG		0x02
#define SPI_AAI1	0xaf
#define SPI_AAI2	0xad
#define SPI_RDSR	0x05
#define SPI_EWSR	0x50
#define SPI_WREN	0x06
#define SPI_WRSR	0x01
#define SPI_WRDI	0x04
#define SPI_RDID_15	0x15
#define SPI_RDID_90	0x90
#define SPI_RDID_JEDEC	0x9f
#define SPI_RDID_AB	0xab
#define	SPI_UNPROT	0x39
#define	SPI_PPRGAE	0x82

#define	BUSY_TIMEOUT	500000L

#define	FLAGS_AT26F004	0x0001
#define	FLAGS_AT25DF	0x0002


static char *copyright = "25F 0.1.3  SPI flash module  "
	"(c) Oleg O. Chukaev  ["__DATE__"]";
static word chtype = 0xFFFF;
static CHDEFS *pchd;
static byte pin_wp = 0;


/*
 | 몫祭 ⠭
 */
static void power_off (void) {
	bdport (PRST);
}


/*
 | 祭 ⠭
 */
static int power_on (void) {

	if (chtype == 0xFFFF)
		return (ERC_CHIPUNSP);

	bdport (PFORCE | PWR1 | PAIN);		/* DD1:IOO */
	bdport (PFORCE | PWR2);			/* DD4:OOO */
	dlp->voltage (dlp->Vcc, 0, 0);
	bdport (PWB2 | 0x80);			/* Vcc pin 36 */
	bdport (PWA2 | BITTER);
	bdport (PWB1 | PIN_HOLD | pin_wp | PIN_CE);
	bdport (PWC1 | PIN_SST45_WP);		/* 1  ࠧ襭  */
	dlp->mcsdelay (10000);
	return (0);
}


/*
 | Get value of one pin
 |
 | Ret: 0 / 1
 */
static byte get_pin (byte pin) {
	return (bdport (PRA1) & pin) ? 1 : 0;
}


/*
 | Set all pins
 */
static void set_pins (byte pins) {
	bdport (PWB1 | pins);
}


static void pin_ce_1 (void) {
  	set_pins (PIN_HOLD | pin_wp | PIN_CE);
}


static void pin_ce_0 (void) {
  	set_pins (PIN_HOLD | pin_wp);
}


/*
 |  
 */
static byte read_spi_byte (void) {

	word i;
	byte d = 0;

	for (i = 0; i < 8; i++) {
		d <<= 1;
		if (cfg_rd_alg == 0) {
			/*  設⢠  訩    SO
			 * 񫪨  ᯠ ⮣   SCK,
			 *  񫪭 訩    SI.
			 */
			d |= get_pin (PIN_SO);
		}
		set_pins (pin_wp | PIN_HOLD | PIN_SCK);	/* SCK=1 */
		set_pins (pin_wp | PIN_HOLD);		/* SCK=0 */
		if (cfg_rd_alg != 0) {
			/*   SST45*   ⠪,   
			 * 񫪨 _᫥騬_ ᮬ  SCK.
			 */
			d |= get_pin (PIN_SO);
		}
	}
	return d;
}


/*
 | । 
 */
static int write_spi_byte (byte d) {

	word i;
	byte z;

	for (i = 0; i < 8; i++) {
		z = (d & 0x80) ? PIN_SI : 0;
		set_pins (pin_wp | PIN_HOLD | z);		/* data */
		set_pins (pin_wp | PIN_HOLD | PIN_SCK | z);	/* SCK=1 */
		set_pins (pin_wp | PIN_HOLD);			/* SCK=0 */
		d <<= 1;
	}
	return 0;
}


/*
 | Unprotect blocks
 */
static void unprotect_all (void) {

	pin_ce_0 ();
	if (cfg_wrsr_alg == 0) {
		write_spi_byte (SPI_EWSR);
	} else { //if (cfg_wrsr_alg == 1) {
		write_spi_byte (SPI_WREN);
	}
	pin_ce_1 ();
	dlp->mcsdelay (1000);
	pin_ce_0 ();
	write_spi_byte (SPI_WRSR);
	write_spi_byte (0x00);		/* clear BP0, BP1, ... */
	pin_ce_1 ();
	dlp->mcsdelay (150000L);	/* for Atmel, Spansion and PMC */
}


/*
 | Unprotect blocks for AT26F004
 */
static void unprotect_at26f004 (void) {

	dword addr;

	/* Clear SPRL bit in the Status Register */
	pin_ce_0 ();
	write_spi_byte (SPI_WREN);
	pin_ce_1 ();
	pin_ce_0 ();
	write_spi_byte (SPI_WRSR);
	write_spi_byte (0x00);
	pin_ce_1 ();
	/* Unprotect 0x00000, 0x10000, ..., 0x70000,
	 * 0x78000, 0x7a000, 0x7c000.
	 */
	for (addr = 0; addr <= 10; addr++) {
		pin_ce_0 ();
		write_spi_byte (SPI_WREN);
		pin_ce_1 ();
		pin_ce_0 ();
		write_spi_byte (SPI_UNPROT);
		write_spi_byte (addr <= 7 ? addr : 0x07);
		write_spi_byte (addr <= 7 ? 0x00 : 0x80 + (addr - 8) * 0x20);
		write_spi_byte (0x00);
		pin_ce_1 ();
	}
}


/*
 | Unprotect blocks
 */
static void unprotect (void) {

	if (cfg_wr_alg == 4)
		return;

	if (cfg_flags & FLAGS_AT26F004) {
		unprotect_at26f004 ();
	} else if (cfg_flags & FLAGS_AT25DF) {
		unprotect_all ();	/* Clear SPRL bit */
		unprotect_all ();	/* Clear bits 5:2 */
	} else {
		unprotect_all ();
	}
}


/*
 | Poll BUSY/WIP bit
 */
static int poll_busy_bit (dword timeout) {

	dword i;

	for (i = 0; i <= timeout / 20; i++) {
		dlp->mcsdelay (20);
		pin_ce_1 ();
//		pin_ce_0 ();
		write_spi_byte (SPI_RDSR);
		if ((read_spi_byte () & 0x01) == 0x00) {
			pin_ce_1 ();
			return 0;
		}
	}
	pin_ce_1 ();
	return 1;
}


/*
 | Read, compare, blank check
 */
static int read_all (void) {

	dword adr;
	dword istep = pchd->size / 64L;
	dword errors = 0;
	byte dat;

	/* Status Register */
	if (pchd->size == 1) {
		power_on ();
		pin_ce_0 ();
		write_spi_byte (SPI_RDSR);
		dat = read_spi_byte ();
		pin_ce_1 ();
		power_off ();
		dlp->setbytebuf (0, dat);
		return 0;
	}

	/* Flash */
	power_on ();
	pin_ce_0 ();
	set_pins (PIN_HOLD);
	if (cfg_rd_alg == 0) {
		/* 25F* */
		write_spi_byte (SPI_READ);
		write_spi_byte (0x00);		/* A23..A16 */
		write_spi_byte (0x00);		/* A15..A8 */
		write_spi_byte (0x00);		/* A7..A0 */
	} else {
		/* SST45* */
		write_spi_byte (0xff);
		write_spi_byte (0x00);		/* A23..A16 */
		write_spi_byte (0x00);		/* A15..A8 */
		write_spi_byte (0x00);		/* A7..A0 */
		write_spi_byte (0xff);
		write_spi_byte (0xff);
	}
	for (adr = 0; adr < pchd->size; adr++) {
		dat = read_spi_byte ();
		switch (dlp->funcode) {
		case DFN_FREAD:
			dlp->setbytebuf (adr, dat);
			break;
		case DFN_FCOMPARE:
			if (dlp->getbytebuf (adr) != dat) {
				/* ⬥ ᮢ */
				dlp->setbmark (adr, 1);
				errors++;
			}
			break;
		case DFN_FBLANK:
			if (dat != 0xff) {
				dlp->adr = adr;
				power_off ();
				return ERC_BLANK;
			}
		}
		if (adr % istep == 0 && dlp->iprogress (adr, 2))
			break;
	}
	power_off ();
	dlp->edata = errors;
	return 0;
}


/*
 | Read ID
 |
 | 0 (JEDEC):	0x9f, [read cont.,] read mfr, read type, read capacity
 |              Eon, ...
 | 1 (Atmel 1):	0x15, read mfr, read dev
 | 2 (Atmel 2):	0xab, read mfr, read dev_h, read dev_l
 | 5 (SST,Eon):	0x90, 0x00, 0x00, 0x00 / 0x01, read mfr / read dev
 |              SST, SST45, Eon
 | 7 (Elan):    0x90, 0x00, 0x00, 0x00, rd 0x7f, rd 0x7f, read mfr, read dev
 | 9 (SST):	0xab, 0x00, 0x00, 0x00, read mfr, read dev[, read mfr2]
 |              SST, [PMC], Eon (mfr == dev == ...)
 */
static int read_id (void) {

	dword id1 = 0, id2 = 0;

	power_on ();
	pin_ce_0 ();
	switch (cfg_id_alg) {
	case 9:
		write_spi_byte (SPI_RDID_AB);
		write_spi_byte (0x00);
		write_spi_byte (0x00);
		write_spi_byte (0x00);		/* ID addr (0: mfr_id first) */
		id1 = read_spi_byte ();
		id2 = read_spi_byte ();
		break;

	case 1:
		write_spi_byte (SPI_RDID_15);
		id1 = read_spi_byte ();
		id2 = read_spi_byte ();
		break;

	case 2:
		write_spi_byte (SPI_RDID_AB);
		id1 = read_spi_byte ();
		id2 = (read_spi_byte () << 8L);
		id2 |= read_spi_byte ();
		break;

	case 0:
		/* JEDEC ID
		 * 0x9f, [read cont.,] read mfr, read type, read capacity
		 */
		write_spi_byte (SPI_RDID_JEDEC);
		id1 = read_spi_byte ();
		if (id1 == 0x7f) {
			id1 = (id1 << 8) | read_spi_byte ();
		}
		id2 = (read_spi_byte () << 8);
		id2 |= read_spi_byte ();
		break;

	case 5:
		write_spi_byte (SPI_RDID_90);
		write_spi_byte (0x00);
		write_spi_byte (0x00);
		write_spi_byte (0x00);
		id1 = read_spi_byte ();
		pin_ce_1 ();
		pin_ce_0 ();
		write_spi_byte (SPI_RDID_90);
		write_spi_byte (0x00);
		write_spi_byte (0x00);
		write_spi_byte (0x01);
		id2 = read_spi_byte ();
		break;

	case 7:
		write_spi_byte (SPI_RDID_90);
		write_spi_byte (0x00);
		write_spi_byte (0x00);
		write_spi_byte (0x00);
		id1 = (read_spi_byte () << 8L);
		id1 |= read_spi_byte ();
		id2 = (read_spi_byte () << 8L);
		id2 |= read_spi_byte ();
	}
	power_off ();
	dlp->edata = id1 | (id2 << 16L);
	return 0;
}


/*
 | Erase
 |
 | 0 (SST, Atmel 2):		0x60
 | 1 (Atmel 1):			0x62
 | 2 (SST 2, Atmel, EON):	0xC7
 | 3 (SST45*):			0x60, 0x??, 0x??, 0x?? 0xD0, 0x??
 */
static int erase (void) {

	dword istep, i, w0;

	if (cfg_er_alg == 3) {
		pin_wp = PIN_WP;
		power_on ();
		pin_ce_0 ();
		write_spi_byte (0x60);
		write_spi_byte (0x00);
		write_spi_byte (0x00);
		write_spi_byte (0x00);
		write_spi_byte (0xD0);
		write_spi_byte (0x00);
		dlp->mcsdelay (cfg_etime);		/* delay */
		power_off ();
		pin_wp = 0;
		return 0;
	}

	pin_wp = PIN_WP;
	power_on ();
	unprotect ();
	pin_ce_0 ();
	write_spi_byte (SPI_WREN);
	pin_ce_1 ();
	pin_ce_0 ();
	if (cfg_er_alg == 0) {
		write_spi_byte (SPI_CERASE);
	} else if (cfg_er_alg == 1) {
		write_spi_byte (SPI_CERASE1);
	} else if (cfg_er_alg == 2) {
		write_spi_byte (SPI_CERASE2);
	}
	pin_ce_1 ();
	if (cfg_etime == 0L) {				/* poll BUSY bit */
		dlp->mcsdelay (10000);
		dlp->iprogress (1600, 0);		/* init, 16 s */
		istep = 1600 / 64;
		i = 0;
		do {
			pin_ce_1 ();
			pin_ce_0 ();
			write_spi_byte (SPI_RDSR);
			w0 = read_spi_byte ();
			dlp->mcsdelay (10000);
			if (i % istep == 0 && dlp->iprogress (i, 2)) {
				break;
			}
			i++;
		} while ((w0 & 0x01) == 0x01);
		pin_ce_1 ();
	} else if (cfg_etime < 300000L) {		/* < 0.3 s */
		dlp->mcsdelay (cfg_etime);		/* delay */
	} else {
		dlp->iprogress (cfg_etime/1000, 0);	/* init */
		istep = cfg_etime / 64000L;
		for (i = 0; i < cfg_etime/1000; i++) {
			if (i % istep == 0 && dlp->iprogress (i, 2)) {
				break;
			}
			dlp->mcsdelay (1000);		/* delay */
		}
	}
	dlp->iprogress (0, 1);				/* close */
	power_off ();
	pin_wp = 0;
	return 0;
}


/*
 | Write
 |
 | AAI/Page mode:
 | 0:		0x02, addr2, addr1, addr0, data0, ..., data128(256)
 | 1 (SST):	0xaf, addr2, addr1, addr0, data0, WAIT, 0xaf, data1, ...
 | 2 (SST):	0xad, addr2, addr1, addr0, data0, data1, WAIT,
 |		0xad, data2, data3, ...
 | 3 (SST45):	0x10, addr2, addr1, addr0, data0, 0x??, WAIT, ...
 | 4 (AT26DF):  0x82, addr2, addr1, addr0, data0, ..., data256
 */
static int write_all (void) {

	dword i;
	dword adr = 0;
	dword istep = pchd->size / 64L;

	/* Status Register */
	if (pchd->size == 1) {
		pin_wp = PIN_WP;
		power_on ();
		pin_ce_0 ();
		write_spi_byte (SPI_WREN);
		pin_ce_1 ();
		dlp->mcsdelay (1000);
		pin_ce_0 ();
		write_spi_byte (SPI_WRSR);
		write_spi_byte (dlp->getbytebuf (0));
		pin_ce_1 ();
		dlp->mcsdelay (150000L);/* for Atmel, Spansion and PMC */
		power_off ();
		pin_wp = 0;
		return 0;
	}

	/* Flash */
	pin_wp = PIN_WP;
	power_on ();
	unprotect ();
	if (cfg_wr_alg != 4) {
		pin_ce_0 ();
		write_spi_byte (SPI_WREN);
		pin_ce_1 ();
	}
	pin_ce_0 ();
	switch (cfg_wr_alg) {
	case 0:
	case 4:
		for (adr = 0; adr < pchd->size; adr += cfg_wpage) {
			if (cfg_wr_alg == 0) {
				write_spi_byte (SPI_PRG);
			} else /* cfg_wr_alg == 4 */ {
				write_spi_byte (SPI_PPRGAE);
			}
			write_spi_byte ((adr >> 16L) & 0xff);	/* A23..A16 */
			write_spi_byte ((adr >> 8L) & 0xff);	/* A15..A8 */
			write_spi_byte (adr & 0xff);		/* A7..A0 */
			for (i = 0; i < cfg_wpage; i++) {
				write_spi_byte (dlp->getbytebuf (adr + i));
			}
			pin_ce_1 ();
//			dlp->mcsdelay (cfg_wtime * cfg_wpage);
			if (poll_busy_bit (cfg_wtime * cfg_wpage)) {
				power_off ();
				pin_wp = 0;
				dlp->adr = adr;
				return ERC_WRITE;
			}
			if (cfg_wr_alg != 4) {
				pin_ce_0 ();
				write_spi_byte (SPI_WREN);
				pin_ce_1 ();
			}
			pin_ce_0 ();
			if (adr % istep == 0 && dlp->iprogress (adr, 2)) {
				break;
			}
		}
		break;

	case 1:
		write_spi_byte (SPI_AAI1);
		write_spi_byte (0x00);		/* A23..A16 */
		write_spi_byte (0x00);		/* A15..A8 */
		write_spi_byte (0x00);		/* A7..A0 */
		write_spi_byte (dlp->getbytebuf (0));
		pin_ce_1 ();
		/* ६   <= 20 ,  ⢥ ,
		 * 祬 ॡ   । 砭  
		 *  ⥭ ॣ ﭨ. ⮬   ,
		 *  横, ᯮ㥬 䨪஢ প.
		 */
		dlp->mcsdelay (cfg_wtime);
		pin_ce_0 ();
		for (adr = 1; adr < pchd->size; adr++) {
			write_spi_byte (SPI_AAI1);
			write_spi_byte (dlp->getbytebuf (adr));
			pin_ce_1 ();
			dlp->mcsdelay (cfg_wtime);
			pin_ce_0 ();
			if (adr % istep == 0 && dlp->iprogress (adr, 2)) {
				break;
			}
		}
		break;

	case 2:
		write_spi_byte (SPI_AAI2);
		write_spi_byte (0x00);		/* A23..A16 */
		write_spi_byte (0x00);		/* A15..A8 */
		write_spi_byte (0x00);		/* A7..A0 */
		write_spi_byte (dlp->getbytebuf (0));
		write_spi_byte (dlp->getbytebuf (1));
		pin_ce_1 ();
		/* ६   <= 10 ,  ⢥ ,
		 * 祬 ॡ   । 砭  
		 *  ⥭ ॣ ﭨ. ⮬   ,
		 *  横, ᯮ㥬 䨪஢ প.
		 */
		dlp->mcsdelay (cfg_wtime);
		pin_ce_0 ();
		for (adr = 2; adr < pchd->size; adr += 2) {
			write_spi_byte (SPI_AAI2);
			write_spi_byte (dlp->getbytebuf (adr));
			write_spi_byte (dlp->getbytebuf (adr+1));
			pin_ce_1 ();
			dlp->mcsdelay (cfg_wtime);
			pin_ce_0 ();
			if (adr % istep == 0 && dlp->iprogress (adr, 2)) {
				break;
			}
		}
		break;

	case 3:
		for (adr = 0; adr < pchd->size; adr++) {
			write_spi_byte (0x10);
			write_spi_byte ((adr >> 16L) & 0xff);	/* A23..A16 */
			write_spi_byte ((adr >> 8L) & 0xff);	/* A15..A8 */
			write_spi_byte (adr & 0xff);		/* A7..A0 */
			write_spi_byte (dlp->getbytebuf (adr));
			write_spi_byte (0xff);
			dlp->mcsdelay (cfg_wtime);
			pin_ce_1 ();
			pin_ce_0 ();
			if (adr % istep == 0 && dlp->iprogress (adr, 2)) {
				break;
			}
		}
		break;
	}

	/*  SST45*   AT26DF* WRDI  ॡ */
	if (cfg_wr_alg != 3 && cfg_wr_alg != 4) {
		write_spi_byte (SPI_WRDI);
	}
	pin_ce_1 ();
	power_off ();
	pin_wp = 0;
	return 0;
}


/*
 | Test
 */
static int test (void) {

	word w1, w2, w3;

	pin_wp = PIN_WP;
	power_on ();

	pin_ce_0 ();
	write_spi_byte (SPI_RDSR);
	w1 = read_spi_byte ();
	pin_ce_1 ();

	unprotect ();

	pin_ce_0 ();
	write_spi_byte (SPI_RDSR);
	w2 = read_spi_byte ();
	pin_ce_1 ();

	pin_ce_0 ();
	write_spi_byte (SPI_WREN);
	pin_ce_1 ();

	pin_ce_0 ();
	write_spi_byte (SPI_RDSR);
	w3 = read_spi_byte ();
	pin_ce_1 ();

	power_off ();
	dlp->msgBox (dmfInformation | dmfOKButton,
		"\003%02X : %02X : %02X", w1, w2, w3);
	pin_wp = 0;
	return 0;
}


/*
 | Lock
 */
static int lock_chip (void) {

	word w1, w2, w3;

	pin_wp = PIN_WP;
	power_on ();

	pin_ce_0 ();
	write_spi_byte (SPI_RDSR);
	w1 = read_spi_byte ();
	pin_ce_1 ();

	pin_ce_0 ();
	if (cfg_wrsr_alg == 0) {
		write_spi_byte (SPI_EWSR);
	} else { //if (cfg_wrsr_alg == 1) {
		write_spi_byte (SPI_WREN);
	}
	pin_ce_1 ();
	dlp->mcsdelay (1000);
	pin_ce_0 ();
	write_spi_byte (SPI_WRSR);
	write_spi_byte (0x0c);		/* clear BP0, BP1, ... */
	pin_ce_1 ();
	dlp->mcsdelay (150000L);	/* for Atmel, Spansion and PMC */

	pin_ce_0 ();
	write_spi_byte (SPI_RDSR);
	w2 = read_spi_byte ();
	pin_ce_1 ();

	pin_ce_0 ();
	write_spi_byte (SPI_WREN);
	pin_ce_1 ();

	pin_ce_0 ();
	write_spi_byte (SPI_RDSR);
	w3 = read_spi_byte ();
	pin_ce_1 ();

	power_off ();
	dlp->msgBox (dmfInformation | dmfOKButton,
		"\003%02X : %02X : %02X", w1, w2, w3);
	pin_wp = 0;
	return 0;
}


/*
 | ⠭ ⨯ ROM
 */
static word settype (void) {

	word i;

	for (i = 0; i < sizeof(chd) / sizeof(CHDEFS); i++) {
		if (chd[i].size == dlp->esize) {
			if (chd[i].size == 1) {
				dlp->funmask[DFN_FCOMPARE] = 0;
				dlp->funmask[DFN_READID] = 0;
				dlp->funmask[DFN_ERASE] = 0;
				dlp->funmask[DFN_FBLANK] = 0;
				dlp->funmask[DFN_TEST] = 0;
				dlp->funmask[DFN_LOCK] = 0;
			} else {
				dlp->funmask[DFN_FCOMPARE] = 1;
				dlp->funmask[DFN_READID] = 1;
				dlp->funmask[DFN_ERASE] = 1;
				dlp->funmask[DFN_FBLANK] = 1;
				dlp->funmask[DFN_TEST] = 1;
				dlp->funmask[DFN_LOCK] = 1;
			}
			if (cfg_wr_alg == 4) {
				dlp->funmask[DFN_ERASE] = 0;
			}
			return (chtype = i);
		}
	}
	return (chtype = 0xFFFF);
}


/*
 | Main code
 */
int dmain (void) {

	switch (dlp->funcode) {
	case DFN_LOAD:
		/* ⨢஢ 㯭 㭪樨: */
		dlp->funmask[DFN_INIT] = 1;
		dlp->funmask[DFN_FREAD] = 1;
		dlp->funmask[DFN_FWRITE] = 1;
		dlp->funmask[DFN_FCOMPARE] = 1;
		dlp->funmask[DFN_READID] = 1;
		dlp->funmask[DFN_ERASE] = 1;
		dlp->funmask[DFN_FBLANK] = 1;
		dlp->funmask[DFN_TEST] = 1;
		dlp->funmask[DFN_LOCK] = 1;
		dlp->defstate = 0xFF;
		dlp->copyright = (byte *)copyright;
		dlp->version = DLLVERSION;
		dlp->cfgkeys = cfgkeys;
		return 0;

	case DFN_INIT:
		if (settype () == 0xFFFF)
			return ERC_CHIPUNSP;
		pchd = &chd[chtype];
		dlp->dippins = 8;
		return 0;

	case DFN_FREAD:
		return read_all ();

	case DFN_FWRITE:
		return write_all ();

	case DFN_FCOMPARE:
		return read_all ();

	case DFN_FBLANK:
		return read_all ();

	case DFN_READID:
		return read_id ();

	case DFN_ERASE:
		return erase ();

	case DFN_TEST:
		return test ();

	case DFN_LOCK:
		return lock_chip ();

	default:
		return ERC_FUNUNSP;
	}
}

/* E0F */
